from riot_na.data.model import AirrRearrangementEntryAA, ShortRegion


def get_region_position_indices(airr: AirrRearrangementEntryAA, region: ShortRegion) -> list[int]:
    match region:
        case ShortRegion.CDR1:
            assert airr.cdr1_start_aa and airr.cdr1_end_aa
            return list(range(airr.cdr1_start_aa - 1, airr.cdr1_end_aa))
        case ShortRegion.CDR2:
            assert airr.cdr2_start_aa and airr.cdr2_end_aa
            return list(range(airr.cdr2_start_aa - 1, airr.cdr2_end_aa))
        case ShortRegion.CDR3:
            assert airr.cdr3_start_aa and airr.cdr3_end_aa
            return list(range(airr.cdr3_start_aa - 1, airr.cdr3_end_aa))
        case ShortRegion.FW1:
            assert airr.fwr1_start_aa and airr.fwr1_end_aa
            return list(range(airr.fwr1_start_aa - 1, airr.fwr1_end_aa))
        case ShortRegion.FW2:
            assert airr.fwr2_start_aa and airr.fwr2_end_aa
            return list(range(airr.fwr2_start_aa - 1, airr.fwr2_end_aa))
        case ShortRegion.FW3:
            assert airr.fwr3_start_aa and airr.fwr3_end_aa
            return list(range(airr.fwr3_start_aa - 1, airr.fwr3_end_aa))
        case ShortRegion.FW4:
            assert airr.fwr4_start_aa and airr.fwr4_end_aa
            return list(range(airr.fwr4_start_aa - 1, airr.fwr4_end_aa))


def get_regions_position_indices(airr: AirrRearrangementEntryAA, regions: list[ShortRegion]) -> list[int]:
    indices = set()
    for region in regions:
        indices.update(get_region_position_indices(airr, region))
    return sorted(list(indices))


def scheme_positions_to_index(airr: AirrRearrangementEntryAA, scheme_positions: list[str]) -> list[int]:
    assert airr.positional_scheme_mapping
    scheme_positional_mapping = {v: k for k, v in airr.positional_scheme_mapping.items()}
    return [scheme_positional_mapping[pos] for pos in scheme_positions]


def get_primary_seq(airr: AirrRearrangementEntryAA) -> str:
    assert airr.scheme_residue_mapping is not None
    return "".join(airr.scheme_residue_mapping.values())



from pathlib import Path
from typing import Optional

from riot_na.alignment.alignment_utils import offset_alignments
from riot_na.alignment.gene_aligner import (
    GeneAligner,
    create_aligner,
    create_v_gene_aligner,
)
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import AlignmentsNT, GermlineGene, Locus, Organism


class VDJCAlignerNT:
    def __init__(
        self,
        v_aligner: GeneAligner,
        d_aligners: dict[Organism, GeneAligner],
        j_aligners: dict[Organism, dict[Locus, GeneAligner]],
        c_aligners: dict[Locus, GeneAligner],
    ) -> None:
        self.v_aligner = v_aligner
        self.d_aligners = d_aligners
        self.j_aligners = j_aligners
        self.c_aligners = c_aligners  # at the time of writing, there is only one c germline database

    def produce_nt_alignments(self, query: str) -> AlignmentsNT:
        result = AlignmentsNT()

        best_v_alignment = self.v_aligner.align(query, both_strains=True)

        if best_v_alignment is None:
            return result

        query_v = best_v_alignment.q_seq
        result.v = best_v_alignment

        # mask alignments
        query_j = best_v_alignment.q_seq[best_v_alignment.q_end :]

        # align j genes
        species = best_v_alignment.species
        locus = best_v_alignment.locus

        j_aligner = self.j_aligners[species][locus]
        best_j_alignment = j_aligner.align(query_j, both_strains=False)

        if best_j_alignment is None:
            return result

        best_j_alignment = offset_alignments(best_v_alignment.q_end, best_j_alignment)
        assert best_j_alignment is not None
        result.j = best_j_alignment

        # mask c input (query_db_clean -> valn_end + jaln_end)
        query_c = query_v[best_j_alignment.q_end :]

        # align c
        c_aligner = self.c_aligners[locus]
        best_c_alignment = c_aligner.align(query_c, both_strains=False)

        if best_c_alignment is not None:
            best_c_alignment = offset_alignments(best_j_alignment.q_end, best_c_alignment)
            result.c = best_c_alignment

        # mask d input (query_db_clean -> valn_end + jaln_end)
        query_d = query_v[best_v_alignment.q_end : best_j_alignment.q_start]

        # align c genes
        d_aligner = self.d_aligners[species]
        best_d_alignment = d_aligner.align(query_d, both_strains=False)

        if best_d_alignment is not None:
            best_d_alignment = offset_alignments(best_v_alignment.q_end, best_d_alignment)
            result.d = best_d_alignment

        return result


def create_vdjc_aligner_nt(allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR):
    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS]

    v_aligner = create_v_gene_aligner(allowed_species=allowed_species, db_dir=db_dir)

    d_aligners = {}
    for organism in allowed_species:
        d_aligner = create_aligner(
            allowed_species=organism, germline_gene=GermlineGene.D, locus=Locus.IGH, db_dir=db_dir
        )
        d_aligners[organism] = d_aligner

    j_aligners: dict[Organism, dict[Locus, GeneAligner]] = {}

    for organism in allowed_species:
        for locus in Locus:
            j_aligner = create_aligner(
                allowed_species=organism, germline_gene=GermlineGene.J, locus=locus, db_dir=db_dir
            )
            organism_aligners = j_aligners.get(organism, {})
            organism_aligners[locus] = j_aligner
            j_aligners[organism] = organism_aligners

    c_aligners = {}

    for locus in Locus:
        c_organism = Organism.HOMO_SAPIENS if organism != Organism.CUSTOM else Organism.CUSTOM

        c_aligner = create_aligner(allowed_species=c_organism, germline_gene=GermlineGene.C, locus=locus, db_dir=db_dir)
        c_aligners[locus] = c_aligner

    vdjc_aligner_nt = VDJCAlignerNT(v_aligner, d_aligners, j_aligners, c_aligners)

    return vdjc_aligner_nt



from collections.abc import Iterator
from dataclasses import replace
from itertools import groupby

from Bio.Seq import Seq

from riot_na.data.model import AlignmentEntry, AlignmentString, Cigar


def get_cigar_op_groups(cigar: Cigar) -> Iterator[tuple[int, str]]:
    cigar_groups = (list(grouper) for _, grouper in groupby(cigar, lambda character: character.isdigit()))
    for group_size_grouper, op_grouper in zip(cigar_groups, cigar_groups):
        yield int("".join(group_size_grouper)), op_grouper[0]


def unfold_cigar(cigar: Cigar) -> AlignmentString:
    result = []
    for group_size, operation in get_cigar_op_groups(cigar):
        for _ in range(group_size):
            result.append(operation)
    return AlignmentString("".join(result))


def fold_cigar(alignment_str: AlignmentString) -> Cigar:
    return Cigar("".join(f"{sum(1 for _ in op_group)}{op}" for op, op_group in groupby(alignment_str)))


def has_frameshift(cigar: Cigar) -> bool:
    counter = 0

    for size, operation in get_cigar_op_groups(cigar):
        if operation == "M":
            if counter % 3 != 0:
                return True
            counter = 0

        if operation == "I":
            counter -= size
        elif operation == "D":
            counter += size

    return False


def translate(query_sequence: str, coding_frame: int) -> str:
    assert coding_frame in [0, 1, 2]

    query_sequence = query_sequence[coding_frame:]
    partial_codon_len = len(query_sequence) % 3
    if partial_codon_len:
        query_sequence = query_sequence[:-partial_codon_len]  # To avoid BioPython "Partial codon" warning.
    coding_dna = Seq(query_sequence)
    return str(coding_dna.translate(gap="."))


def infer_reading_frame(t_start: int, t_frame: int) -> int:
    #             AT GAC TATGCTATGGACTACTGGGGTCAAGGAACCTCAGTCACCGTCTCCTCAG
    #                 MM MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
    #  T TAC TAC TAC TAC TACGGTATGGACGTCTGGGGGCAAGGGACCACGGTCACCGTCTCCTCAG target reading frame is 1
    #    |            |
    #  t_frame     t_start
    # t_start = 11
    # t_frame = 1
    # frame_offset = ((t_start -t_frame) %3) = 10 % 3 = 1
    # q_frame = (3 - frame_offset) %3 = (3 - 1) % 3 = 2

    # offset from the reading frame of the first aligned position
    frame_offset = (t_start - t_frame) % 3
    q_frame = (3 - frame_offset) % 3
    return q_frame


def offset_alignments(offset: int, aln: AlignmentEntry) -> AlignmentEntry:
    return replace(aln, q_start=aln.q_start + offset, q_end=aln.q_end + offset)


def align_sequences(query: str, target: str, alignment: AlignmentString) -> tuple[str, str]:
    query_aln = []
    target_aln = []

    for op in alignment:
        if op == "M":
            query_aln.append(query[0])
            target_aln.append(target[0])
            query = query[1:]
            target = target[1:]
        elif op == "I":
            query_aln.append(query[0])
            target_aln.append("-")
            query = query[1:]
        elif op == "D":
            query_aln.append("-")
            target_aln.append(target[0])
            target = target[1:]

    return "".join(query_aln), "".join(target_aln)



import csv
from dataclasses import fields
from typing import Any, Dict, Iterable, Optional, Type

from riot_na.data.model import AirrRearrangementEntry_co, serialize_airr_entry


class AirrRearrangementEntryWriter:
    def __init__(
        self,
        file_handle: Any,
        cls: Type[AirrRearrangementEntry_co],
        dialect: str = "excel",
        fmtparams: Optional[Dict[str, Any]] = None,
    ):
        if not file_handle:
            raise ValueError("The f argument is required")

        if fmtparams is None:
            fmtparams = {}

        self._fieldnames = [x.name for x in fields(cls)]
        self._writer = csv.DictWriter(file_handle, dialect=dialect, fieldnames=self._fieldnames, **fmtparams)

    def write(self, data: Iterable[AirrRearrangementEntry_co], skip_header: bool = False):
        if not skip_header:
            self._writer.writeheader()

        for item in data:
            row = serialize_airr_entry(item)
            self._writer.writerow(row)



import math
from inspect import getframeinfo

import pandas as pd  # type: ignore

from riot_na.airr.airr_validator import REGIONS
from riot_na.data.model import AirrRearrangementEntryNT


def get_sequence_alignment_with_masked_regions(record: pd.Series) -> str:
    def _get_mask_char(region_name: str) -> str:
        return "_" if "fwr" in region_name else "-"

    seq = record["sequence_alignment"]
    regions = record[REGIONS].replace(math.nan, None).to_dict().items()  # type: ignore
    for region_name, region in regions:
        if region is not None and len(region) > 0:
            seq = seq.replace(region, _get_mask_char(region_name) * len(region))
    return seq


def get_asserion_error_msg(currentframe, rearrangement: AirrRearrangementEntryNT) -> str:
    frameinfo = getframeinfo(currentframe)
    return (
        f"{frameinfo.filename}:{frameinfo.lineno} "
        f"sequence_id: {rearrangement.sequence_header.split(' ', maxsplit=1)[0]}"
    )



import math
from enum import Enum

import blosum  # type: ignore

from riot_na.alignment.alignment_utils import get_cigar_op_groups, unfold_cigar
from riot_na.data.model import Cigar

# L (Lambda) and K constants are depended on scoring matrix and gap penalties
# and are calculated by ALP (Ascending Ladder Program) - https://doi.org/10.1093%2Fbioinformatics%2Fbtv575
# The present values were calculated for the following parameters (default SSW):
# Match score: 2
# Mismatch score:  -3
# Gap open penalty: 5
# Gap extend penalty: 2
# Background probabilities:  A, T, G, C 0.2499975, N 0.00001


class GumbellParams(Enum):
    IGBLAST = {"L": 1.08, "K": 0.28}
    AA = {"L": 0.26453605633241922, "K": 0.043186874595437463}


def compute_raw_score_aa(
    query: str,
    target: str,
    cigar: Cigar,
    gap_open_penalty=11,
    gap_extend_penalty=1,
    substitution_matrix=blosum.BLOSUM(62),
) -> float:
    raw_score = 0
    query_pos = 0
    target_pos = 0
    prev_match = True
    for op in unfold_cigar(cigar):
        if op == "M":
            raw_score += substitution_matrix[query[query_pos]][target[target_pos]]

            query_pos += 1
            target_pos += 1
            prev_match = True
        elif op == "I":
            if prev_match:
                raw_score -= gap_open_penalty
                prev_match = False
            else:
                raw_score -= gap_extend_penalty
            query_pos += 1
        elif op == "D":
            if prev_match:
                raw_score -= gap_open_penalty
                prev_match = False
            else:
                raw_score -= gap_extend_penalty
            target_pos += 1
    return raw_score


def compute_bit_score(raw_score: int, gumbell_params: GumbellParams = GumbellParams.IGBLAST) -> float:
    return (gumbell_params.value["L"] * raw_score - math.log(gumbell_params.value["K"])) / math.log(2)


def compute_raw_score_from_bit_score(bit_score: float, gumbell_params: GumbellParams = GumbellParams.IGBLAST) -> float:
    return (math.log(gumbell_params.value["K"]) + bit_score * math.log(2)) / gumbell_params.value["L"]


def compute_evalue(query_length: int, db_length: int, bit_score: float) -> float:
    return query_length * db_length * 2 ** (-bit_score)


def calculate_seq_identity(cigar: Cigar, query: str, target: str, query_start: int = 0, target_start: int = 0) -> float:
    cigar_items = get_cigar_op_groups(cigar)
    query_pos = query_start
    target_pos = target_start
    match_cnt = 0
    total = 0
    for op_cnt, op in cigar_items:
        match op:
            case "M":
                for i in range(op_cnt):
                    if query[query_pos + i] == target[target_pos + i]:
                        match_cnt += 1
                query_pos += op_cnt
                target_pos += op_cnt
            case "I":
                query_pos += op_cnt
            case "D":
                target_pos += op_cnt
        total += op_cnt
    return match_cnt / total



import os
from functools import partial
from pathlib import Path
from typing import Callable, Optional, Sequence

import blosum  # type: ignore
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from skbio.alignment import StripedSmithWaterman  # type: ignore

from riot_na.alignment.skbio_alignment import (
    align_aa,  # type: ignore  # pylint: disable=import-error
)
from riot_na.alignment.skbio_alignment import align
from riot_na.common.multi_species_prefiltering import MultiSpeciesPrefiltering
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AlignmentEntryAA,
    AlignmentEntryNT,
    Gene,
    GeneAA,
    GermlineGene,
    Locus,
    Organism,
    SpeciesPrefilteringResult,
)

ALIGNER_PARAMS = {"match_score": 1, "mismatch_score": -1, "gap_open_penalty": 4, "gap_extend_penalty": 1}


class GeneAligner:
    def __init__(
        self,
        genes: list[Gene],
        kmer_size: int,
        distance_threshold: int,
        top_n: int,
        modulo_n: int,
        e_value_threshold: Optional[float] = None,
        alignment_length_threshold: Optional[int] = None,
    ) -> None:
        self.prefiltering = MultiSpeciesPrefiltering(genes, kmer_size, distance_threshold, top_n, modulo_n)

        self.gene_lookup = dict(map(lambda gene: (gene.species + "|" + gene.name, gene), genes))
        self.db_length = sum(len(gene.sequence) for gene in genes)

        self.e_value_threshold = e_value_threshold
        self.alignment_length_threshold = alignment_length_threshold

    def _prefilter(self, query: str, both_strains: bool) -> SpeciesPrefilteringResult:
        if both_strains:
            return self.prefiltering.calculate_top_matches_with_rev_comp(query)
        return self.prefiltering.calculate_top_matches(query)

    def _align_sequences(self, query: str, prefiltering_result: SpeciesPrefilteringResult) -> list[AlignmentEntryNT]:
        aligner = StripedSmithWaterman(query, **ALIGNER_PARAMS)
        rev_comp_aligner = StripedSmithWaterman(prefiltering_result.rev_comp_query, **ALIGNER_PARAMS)

        result = []

        for gene_match in prefiltering_result.top_matches:
            current_species = gene_match.species
            match gene_match.rev_comp:
                case False:
                    alignment = align(
                        aligner=aligner,
                        target_id=gene_match.gene_id,
                        target=self.gene_lookup[gene_match.species + "|" + gene_match.gene_id].sequence,
                        query=query,
                        db_length=self.db_length,
                        rev_comp=gene_match.rev_comp,
                    )
                case True:
                    alignment = align(
                        aligner=rev_comp_aligner,
                        target_id=gene_match.gene_id,
                        target=self.gene_lookup[gene_match.species + "|" + gene_match.gene_id].sequence,
                        query=prefiltering_result.rev_comp_query,
                        db_length=self.db_length,
                        rev_comp=gene_match.rev_comp,
                    )

            alignment_length = alignment.t_end - alignment.t_start

            if self.e_value_threshold is not None:
                assert alignment.e_value is not None
                if alignment.e_value > self.e_value_threshold:
                    continue
            if self.alignment_length_threshold is not None:
                if alignment_length < self.alignment_length_threshold:
                    continue

            target_gene: Gene = self.gene_lookup[gene_match.species + "|" + gene_match.gene_id]

            result.append(
                AlignmentEntryNT(
                    target_id=alignment.target_id,
                    alignment_score=alignment.alignment_score,
                    seq_identity=alignment.seq_identity,
                    e_value=alignment.e_value,
                    q_start=alignment.q_start,
                    q_end=alignment.q_end,
                    q_len=alignment.q_len,
                    t_start=alignment.t_start,
                    t_end=alignment.t_end,
                    t_len=alignment.t_len,
                    cigar=alignment.cigar,
                    rev_comp=alignment.rev_comp,
                    species=current_species,
                    locus=gene_match.locus,
                    q_seq=alignment.query,
                    t_seq=target_gene.sequence,
                    reading_frame=target_gene.reading_frame,
                )
            )

        return result

    def align(self, query: str, both_strains: bool) -> Optional[AlignmentEntryNT]:
        prefiltering_result = self._prefilter(query, both_strains)
        alignments = self._align_sequences(query, prefiltering_result)

        alignments.sort()

        return alignments[0] if len(alignments) > 0 else None


AA_ALIGNER_PARAMS = {
    "gap_open_penalty": 11,
    "gap_extend_penalty": 1,
    "protein": True,
    "substitution_matrix": blosum.BLOSUM(62),
}


class GeneAlignerAA:
    def __init__(
        self,
        genes: list[Gene],
        kmer_size: int,
        distance_threshold: int,
        top_n: int,
        modulo_n: int,
        e_value_threshold: Optional[float] = None,
        alignment_length_threshold: Optional[int] = None,
        max_cdr3_length: Optional[int] = None,
    ) -> None:
        self.prefiltering = MultiSpeciesPrefiltering(genes, kmer_size, distance_threshold, top_n, modulo_n)

        self.gene_lookup: dict[str, Gene] = dict(map(lambda gene: (gene.species + "|" + gene.name, gene), genes))
        self.db_length = sum(len(gene.sequence) for gene in genes)

        self.e_value_threshold = e_value_threshold
        self.alignment_length_threshold = alignment_length_threshold
        self.max_cdr3_length = max_cdr3_length

    def _prefilter(self, query: str) -> SpeciesPrefilteringResult:
        return self.prefiltering.calculate_top_matches(query)

    def _align_sequences(self, query: str, prefiltering_result: SpeciesPrefilteringResult) -> list[AlignmentEntryAA]:
        aligner = StripedSmithWaterman(query, **AA_ALIGNER_PARAMS)

        result = []

        for gene_match in prefiltering_result.top_matches:
            current_species = gene_match.species
            alignment = align_aa(
                aligner=aligner,
                target_id=gene_match.gene_id,
                target=self.gene_lookup[gene_match.species + "|" + gene_match.gene_id].sequence,
                query=query,
                db_length=self.db_length,
            )

            alignment_length = alignment.t_end - alignment.t_start

            if self.e_value_threshold is not None:
                assert alignment.e_value is not None
                if alignment.e_value > self.e_value_threshold:
                    continue
            if self.alignment_length_threshold is not None:
                if alignment_length < self.alignment_length_threshold:
                    continue
            if self.max_cdr3_length is not None:
                if alignment.q_start > self.max_cdr3_length:
                    continue

            target_gene: Gene = self.gene_lookup[gene_match.species + "|" + gene_match.gene_id]

            result.append(
                AlignmentEntryAA(
                    target_id=alignment.target_id,
                    alignment_score=alignment.alignment_score,
                    seq_identity=alignment.seq_identity,
                    e_value=alignment.e_value,
                    q_start=alignment.q_start,
                    q_end=alignment.q_end,
                    t_start=alignment.t_start,
                    t_end=alignment.t_end,
                    cigar=alignment.cigar,
                    species=current_species,
                    locus=gene_match.locus,
                    q_seq=query,
                    t_seq=target_gene.sequence,
                )
            )

        return result

    def align(self, query: str) -> Optional[AlignmentEntryAA]:
        prefiltering_result = self._prefilter(query)
        alignments = self._align_sequences(query, prefiltering_result)

        alignments.sort()

        return alignments[0] if len(alignments) > 0 else None


def get_aligner_params(germline_gene: GermlineGene, locus: Optional[Locus]) -> dict:
    # ideally this should be read from a file or env

    match germline_gene:
        case GermlineGene.V:
            return {
                "kmer_size": 9,
                "distance_threshold": 13,
                "top_n": int(os.environ.get("TOP_N", 12)),
                "modulo_n": 2,
                "e_value_threshold": 0.001,
                "alignment_length_threshold": 100,
            }
        case GermlineGene.D:
            return {
                "kmer_size": 5,
                "distance_threshold": 3,
                "top_n": int(os.environ.get("TOP_N", 5)),
                "modulo_n": 1,
            }
        case GermlineGene.J:
            assert locus is not None
            match locus:
                case Locus.IGH:
                    return {
                        "kmer_size": 5,
                        "distance_threshold": 3,
                        "top_n": int(os.environ.get("TOP_N", 5)),
                        "modulo_n": 2,
                    }
                case Locus.IGK:
                    return {
                        "kmer_size": 5,
                        "distance_threshold": 5,
                        "top_n": int(os.environ.get("TOP_N", 5)),
                        "modulo_n": 1,
                    }
                case Locus.IGL:
                    return {
                        "kmer_size": 5,
                        "distance_threshold": 7,
                        "top_n": int(os.environ.get("TOP_N", 5)),
                        "modulo_n": 2,
                    }
        case GermlineGene.C:
            assert locus is not None
            match locus:
                case Locus.IGH:
                    return {
                        "kmer_size": 5,
                        "distance_threshold": 5,
                        "top_n": int(os.environ.get("TOP_N", 5)),
                        "modulo_n": 2,
                    }
                case Locus.IGK:
                    return {
                        "kmer_size": 3,
                        "distance_threshold": 5,
                        "top_n": int(os.environ.get("TOP_N", 3)),
                        "modulo_n": 1,
                    }
                case Locus.IGL:
                    return {
                        "kmer_size": 3,
                        "distance_threshold": 5,
                        "top_n": int(os.environ.get("TOP_N", 3)),
                        "modulo_n": 1,
                    }


def get_aa_aligner_params(germline_gene: GermlineGene) -> dict:
    # ideally this should be read from a file or env

    match germline_gene:
        case GermlineGene.V:
            return {
                "kmer_size": 3,
                "distance_threshold": 3,
                "top_n": int(os.environ.get("TOP_N", 20)),
                "modulo_n": 1,
                "e_value_threshold": 1e-55,
                "alignment_length_threshold": 80,
            }
        case GermlineGene.J:
            return {
                "kmer_size": 3,
                "distance_threshold": 3,
                "top_n": int(os.environ.get("TOP_N", 5)),
                "modulo_n": 1,
                "alignment_length_threshold": 7,
                "max_cdr3_length": 60,
            }
        case _:
            raise ValueError("AA aligner not supported for gene type")


def get_gene_parsing_function(gene_type: GermlineGene) -> Callable[[str, SeqRecord], Gene]:
    match gene_type:
        case GermlineGene.V:
            return lambda description, record: Gene(
                name=description[0],
                locus=Locus[description[1]],
                reading_frame=int(description[2]),
                species=Organism[description[3]],
                sequence=str(record.seq),
            )
        case GermlineGene.D:
            return lambda description, record: Gene(
                name=description[0],
                locus=Locus.IGH,
                reading_frame=None,
                species=Organism[description[1]],
                sequence=str(record.seq),
            )
        case GermlineGene.J:
            return lambda description, record: Gene(
                name=description[0],
                locus=Locus[description[1]],
                reading_frame=int(description[2]),
                species=Organism[description[3]],
                sequence=str(record.seq),
            )
        case GermlineGene.C:
            return lambda description, record: Gene(
                name=description[0],
                locus=Locus[description[1]],
                reading_frame=None,
                species=Organism.HOMO_SAPIENS,
                sequence=str(record.seq),
            )


def parse_gene_aa(description: Sequence[str], record: SeqRecord, species: Organism):
    return GeneAA(name=description[0], locus=Locus[description[1]], species=species, sequence=str(record.seq))


def read_genes(path: Path, parsing_fn):
    result = []
    for record in SeqIO.parse(path, "fasta"):
        description = record.description.split("\t")
        gene = parsing_fn(description, record)

        result.append(gene)

    return result


def create_v_gene_aligner(allowed_species: list[Organism], db_dir: Path = GENE_DB_DIR) -> GeneAligner:
    genes = []

    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS]

    gene_parsing_function = get_gene_parsing_function(GermlineGene.V)

    for species in allowed_species:
        input_path = Path(db_dir) / "gene_db" / "v_genes" / f"{species.value}.fasta"
        species_genes = read_genes(input_path, gene_parsing_function)

        genes.extend(species_genes)

    aligner_params = get_aligner_params(GermlineGene.V, None)
    genes_aligner = GeneAligner(genes, **aligner_params)

    return genes_aligner


def create_aligner(
    allowed_species: Organism, germline_gene: GermlineGene, locus: Locus, db_dir: Path = GENE_DB_DIR
) -> GeneAligner:
    assert germline_gene != GermlineGene.V

    gene_parsing_function = get_gene_parsing_function(germline_gene)

    gene_path = (
        Path(db_dir)
        / "gene_db"
        / f"{germline_gene.value.lower()}_genes"
        / allowed_species.value
        / f"{locus.value}.fasta"
    )
    genes = read_genes(gene_path, gene_parsing_function)

    aligner_params = get_aligner_params(germline_gene, locus)
    genes_aligner = GeneAligner(genes, **aligner_params)

    return genes_aligner


def create_aa_v_gene_aligner(allowed_species: list[Organism], aa_genes_dir: Path) -> GeneAlignerAA:
    genes = []

    if not allowed_species:
        allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS]

    for species in allowed_species:
        input_path = aa_genes_dir / "v_genes" / f"{species.value}.fasta"

        assert input_path.exists(), f"Input germline file: {str(input_path)} does not exists"

        species_genes = read_genes(input_path, partial(parse_gene_aa, species=species))

        genes.extend(species_genes)

    aligner_params = get_aa_aligner_params(GermlineGene.V)
    genes_aligner = GeneAlignerAA(genes, **aligner_params)

    return genes_aligner


def create_aa_j_gene_aligner(organism: Organism, locus: Locus, aa_genes_dir: Path) -> GeneAlignerAA:
    input_path = aa_genes_dir / "j_genes" / organism.value / f"{locus.value}.fasta"
    assert input_path.exists(), f"Input germline file: {str(input_path)} does not exists"

    genes = read_genes(input_path, partial(parse_gene_aa, species=organism))

    aligner_params = get_aa_aligner_params(GermlineGene.J)
    genes_aligner = GeneAlignerAA(genes, **aligner_params)

    return genes_aligner



# pylint: disable=too-many-branches, too-many-statements

from pathlib import Path
from typing import Optional

from riot_na.airr.airr_builder import AirrBuilder
from riot_na.airr.airr_builder_aa import AirrBuilderAA
from riot_na.alignment.aa_gene_alignments import (
    VJAlignerAA,
    VJAlignmentTranslatorAA,
    create_vj_aligner_aa,
    create_vj_alignment_translator_aa,
)
from riot_na.alignment.nt_gene_alignments import VDJCAlignerNT, create_vdjc_aligner_nt
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AirrRearrangementEntryAA,
    AirrRearrangementEntryNT,
    Organism,
    Scheme,
)
from riot_na.schemes.region_offsets import (
    infer_aa_region_offsets,
    infer_nt_region_offsets,
    infer_region_offsets,
)
from riot_na.schemes.scheme_alignment import (
    SchemeAligner,
    fix_imgt_cdrs_numbering,
    get_positional_scheme_mapping,
    get_scheme_residue_mapping,
    produce_numbering,
    sort_imgt_numbering,
)


class RiotNumberingNT:
    def __init__(
        self,
        vdjc_aligner_nt: VDJCAlignerNT,
        vj_alignment_translator_aa: VJAlignmentTranslatorAA,
        scheme_aligner: SchemeAligner,
    ):
        self.vdjc_aligner = vdjc_aligner_nt
        self.vj_alignment_translator_aa = vj_alignment_translator_aa
        self.scheme_aligner = scheme_aligner

    def run_on_sequence(
        self,
        header: str,
        query_sequence: str,
        scheme: Scheme = Scheme.IMGT,
    ) -> AirrRearrangementEntryNT:  # pylint: disable=too-many-statements
        """
        Align and number nucleotide sequence.

        Parameters:
        -----------
        - header : Fasta sequence header, usually this is simply sequence ID.
        - query_sequence : Nucleotide sequence to number.
        - scheme : Which numbering scheme to use (default IMGT).

        Returns:
        -----------
        AirrRearrangementEntryNT object with alignment results (Extended AIRR format)
        """
        query_sequence = query_sequence.upper()
        translated_alignments = None
        nt_offsets = None
        aa_offsets = None
        sch_alignment = None
        positional_scheme_mapping = None
        numbering = None
        scheme_alignment_exc = None

        # produce alignments for v,d,j,c
        alignments = self.vdjc_aligner.produce_nt_alignments(query_sequence)

        # Query might have been reverse complemented during V alignment.
        if alignments.v:
            query_sequence = alignments.v.q_seq
        try:
            translated_alignments = self.vj_alignment_translator_aa.translate_nt_alignments(alignments)

            if translated_alignments:
                sch_alignment = self.scheme_aligner.align_to_scheme(translated_alignments.aa_alignments, scheme)

                # calculate aa offsets based on the numbering (iterate over cigar to produce positional-> imgt mapping)
                assert sch_alignment
                assert alignments.v
                locus = alignments.v.locus

                region_offsets = infer_region_offsets(sch_alignment, scheme, locus)
                aa_offsets = infer_aa_region_offsets(region_offsets)

                alignment_start = alignments.v.q_start

                nt_offsets = infer_nt_region_offsets(
                    region_offsets, alignment_start, translated_alignments.reading_frame
                )
                numbering = produce_numbering(
                    translated_alignments.translated_query, sch_alignment.alignment_str, sch_alignment.q_start
                )

                if scheme == Scheme.IMGT:
                    numbering = fix_imgt_cdrs_numbering(numbering)
                    numbering = sort_imgt_numbering(numbering)

                scheme_residue_mapping = get_scheme_residue_mapping(numbering)
                positional_scheme_mapping = get_positional_scheme_mapping(scheme_residue_mapping, sch_alignment.q_start)
        except AssertionError as err:
            scheme_alignment_exc = str(err)

        # produce airr result
        # this is separated on purpose to enable other output formats
        airr_builder = AirrBuilder(header, query_sequence, scheme)

        if alignments.v:
            airr_builder.with_v_gene_alignment(alignments.v)

        if translated_alignments:
            airr_builder.with_aa_alignments(translated_alignments)

        if alignments.j:
            airr_builder.with_j_gene_alignment(alignments.j)

        if alignments.d:
            airr_builder.with_d_gene_alignment(alignments.d)

        if alignments.c:
            airr_builder.with_c_gene_alignment(alignments.c)

        if sch_alignment:
            assert nt_offsets
            assert aa_offsets
            airr_builder.with_nt_region_offsets(nt_offsets)
            airr_builder.with_aa_region_offsets(aa_offsets)
            airr_builder.with_aa_scheme_alignment(sch_alignment)

        if numbering:
            (
                airr_builder.with_scheme_residue_mapping(scheme_residue_mapping).with_positional_scheme_mapping(
                    positional_scheme_mapping
                )
            )

        if scheme_alignment_exc:
            airr_builder.with_scheme_alignment_exc(scheme_alignment_exc)

        result = airr_builder.build()
        return result


class RiotNumberingAA:
    def __init__(self, vj_aligner_aa: VJAlignerAA, scheme_aligner: SchemeAligner):
        self.vj_aligner_aa = vj_aligner_aa
        self.scheme_aligner = scheme_aligner

    def run_on_sequence(
        self,
        header: str,
        query_sequence: str,
        scheme: Scheme = Scheme.IMGT,
        extend_alignment: bool = True,
    ) -> AirrRearrangementEntryAA:  # pylint: disable=too-many-statements
        """
        Align and number amino acid sequence.

        Parameters:
        -----------
        - header : Fasta sequence header, usually this is simply sequence ID.
        - query_sequence : Nucleotide sequence to number.
        - scheme : Which numbering scheme to use (default IMGT).
        - extend_alignment: RIOT uses Striped Smith-Waterman local alignemtn algorithm.
        Due to that, if query sequence has too many differences (due to mutations or other reasons)
        to the germline sequence at 5' or / and 3' ends the result may not show that part of the gene.
        This option appends missing amino acids, so the result covers full gene sequence. Beware that
        those query fragments are not truly a part of alignemnt and are shown only for convenience!

        Returns:
        -----------
        AirrRearrangementEntryAA object with alignment results (Extended AIRR format)
        """
        query_sequence = query_sequence.upper()
        sch_alignment = None
        aa_offsets = None
        aligned_sequence = None
        positional_scheme_mapping = None
        numbering = None
        scheme_alignment_exc = None

        # produce alignments for v,d,j,c
        alignments = self.vj_aligner_aa.produce_aa_alignments(query_sequence)

        if alignments.v:
            try:
                sch_alignment = self.scheme_aligner.align_to_scheme(
                    alignments, scheme, extend_alignment=extend_alignment
                )

                # calculate aa offsets based on the numbering (iterate over cigar to produce positional-> imgt mapping)
                assert sch_alignment
                locus = alignments.v.locus

                region_offsets = infer_region_offsets(sch_alignment, scheme, locus)
                aa_offsets = infer_aa_region_offsets(region_offsets)

                alignment_start = sch_alignment.q_start
                alignment_end = sch_alignment.q_end

                aligned_sequence = query_sequence[alignment_start:alignment_end]

                numbering = produce_numbering(aligned_sequence, sch_alignment.alignment_str)

                if scheme == Scheme.IMGT:
                    numbering = fix_imgt_cdrs_numbering(numbering)
                    numbering = sort_imgt_numbering(numbering)

                scheme_residue_mapping = get_scheme_residue_mapping(numbering)
                positional_scheme_mapping = get_positional_scheme_mapping(scheme_residue_mapping, sch_alignment.q_start)
            except AssertionError as err:
                scheme_alignment_exc = str(err)

        # produce airr result
        # this is separated on purpose to enable other output formats
        airr_builder = AirrBuilderAA(header, query_sequence, scheme)

        if alignments.v:
            airr_builder.with_v_gene_alignment_aa(alignments.v)

        if alignments.j:
            airr_builder.with_j_gene_alignment_aa(alignments.j)

        if sch_alignment:
            assert aa_offsets
            airr_builder.with_aa_region_offsets(aa_offsets)
            airr_builder.with_aa_scheme_alignment(sch_alignment)

        if numbering:
            (
                airr_builder.with_scheme_residue_mapping(scheme_residue_mapping).with_positional_scheme_mapping(
                    positional_scheme_mapping
                )
            )

        if scheme_alignment_exc:
            airr_builder.with_scheme_alignment_exc(scheme_alignment_exc)

        result = airr_builder.build()
        return result


def create_riot_nt(
    allowed_species: Optional[list[Organism]] = None,
    db_dir: Path = GENE_DB_DIR,
) -> RiotNumberingNT:
    """
    Factory function for creating RiotNumberingNT object.

    Parameters:
    -----------
    - allowed_species : Limit gene database to specified Organism list (default use all species - mouse and human).
    - allowed_schemes : Limit numbering capabilities to the list of specified schemes (default use all schemes).
    - db_dir : Path to gene and scheme mappings database directory (default use embeded database).

    Returns:
    -----------
    RiotNumberingNT object used for performing alignment and numbering.
    """
    vdjc_aligner_nt = create_vdjc_aligner_nt(allowed_species=allowed_species, db_dir=db_dir)
    vj_alignment_translator_aa = create_vj_alignment_translator_aa(allowed_species=allowed_species, db_dir=db_dir)
    scheme_aligner = SchemeAligner(allowed_species=allowed_species, db_dir=db_dir)
    return RiotNumberingNT(vdjc_aligner_nt, vj_alignment_translator_aa, scheme_aligner)


def create_riot_aa(
    allowed_species: Optional[list[Organism]] = None,
    db_dir: Path = GENE_DB_DIR,
) -> RiotNumberingAA:
    """
    Factory function for creating RiotNumberingNT object.

    Parameters:
    -----------
    - allowed_species : Limit gene database to specified Organism list (default use all species - mouse and human).
    - allowed_schemes : Limit numbering capabilities to the list of specified schemes (default use all schemes).
    - db_dir : Path to gene and scheme mappings database directory (default use embeded database).

    Returns:
    -----------
    RiotNumberingAA object used for performing alignment and numbering.
    """
    vj_aligner_aa = create_vj_aligner_aa(
        allowed_species=allowed_species,
        db_dir=db_dir,
    )
    scheme_aligner = SchemeAligner(allowed_species=allowed_species, db_dir=db_dir)
    return RiotNumberingAA(vj_aligner_aa, scheme_aligner)


if __name__ == "__main__":
    import json
    from dataclasses import asdict

    vdj_alnr = create_vdjc_aligner_nt()
    vdj_aa_translator = create_vj_alignment_translator_aa()
    scheme_alnr = SchemeAligner()

    # riot_numbering = RiotNumberingNT(vdj_alnr, vdj_aa_translator, scheme_alnr)
    # # given query
    # QUERY = "CAGGTGCAGCTACAGCAGTGGGGCGCAGGACTGTTGAAGCCTTCGGAGACCCTGTCCCTCACCTGCGCTGTCTATGGTGGGTCCTTCAGTGGTTACTACTGGAGCTGGATCCGCCAGCCCCCAGGGAAGGGGCTGGAGTGGATTGGGGAAATCAATCATAGTGGAAGCACCAACTACAACCCGTCCCTCAAGAGTCGAGTCACCATATCAGTAGACACGTCCAAGAACCAGTTCTCCCTGAAGCTGAGCTCTGTGACCGCCGCGGACACGGCTGTGTATTACTGTGCGAGAGAAAGGCGAGGTTGTAGTAGTACCAGCTGCTTTTGGATTACTCAACGTGGATACAGCTATGGAGGCTCCTACTACTACTACTACATGGACGTCTGGGGCAAAGGGACCACGGTCACCGTCTCCTCA"
    # sample_result = riot_numbering.run_on_sequence("header", QUERY, Scheme.IMGT)
    # print(json.dumps(asdict(sample_result), indent=4))

    vdj_aa_alnr = create_vj_aligner_aa()
    aa_numbering = RiotNumberingAA(vdj_aa_alnr, scheme_alnr)

    AA_QUERY = "QLQLQESGPGLVKPSETLSLTCTVSGDSISSGDYYWGWIRQPPGKGLEWIGHIYYSGATYYNPSLENRVTISVDTSKNQFSLKLSSVTAADTAVYYCTRDDSSNWRSRGQGTLVTVSS"

    aa_sample_result = aa_numbering.run_on_sequence("header", AA_QUERY, Scheme.IMGT)
    print(json.dumps(asdict(aa_sample_result), indent=4))



# pylint: skip-file
# flake8: noqa
import blosum
import pytest
from skbio.alignment import StripedSmithWaterman

from riot_na.alignment.skbio_alignment import align_aa
from riot_na.data.model import AlignmentEntryAA, AlignmentString, Cigar
from riot_na.schemes.scheme_alignment import (
    _merge_cigars,
    fix_imgt_cdrs_numbering,
    force_c_terminus_matches,
    force_n_terminus_matches,
    infer_last_scheme_position_aligned,
    merge_v_j,
)
from riot_na.schemes.smooth_alignment import (
    reorder_buffer_imgt,
    reorder_buffer_other_schemes,
)

AA_ALIGNER_PARAMS = {
    "gap_open_penalty": 11,
    "gap_extend_penalty": 1,
    "protein": True,
    "substitution_matrix": blosum.BLOSUM(62),
}


def test_align_aa():
    # given
    query = "AAAA"
    target = "BAAABBB"
    expected_cigar = Cigar("3M")

    # when
    aligner = StripedSmithWaterman(query, **AA_ALIGNER_PARAMS)
    result = align_aa(aligner, query, "target_id", target, db_length=None, calculate_score=False)

    # then
    assert result.cigar == expected_cigar
    assert result.q_start == 0
    assert result.q_end == 3
    assert result.t_start == 1
    assert result.t_end == 4


def test_force_n_terminus_alignments():
    # given alignment of query and target
    alignment = AlignmentEntryAA(
        target_id="target_id",
        alignment_score=None,
        seq_identity=None,
        e_value=None,
        q_start=3,
        q_end=5,
        t_start=2,
        t_end=4,
        cigar=Cigar("3M"),
        species=None,
        locus=None,
        q_seq=None,
        t_seq=None,
    )

    # when
    result = force_n_terminus_matches(alignment)

    # then
    # verify 1S was appended to cigra (q_start - t_start == 1)
    # verify that no N was appended
    # verify that min(q_start, t_start) matches was appended to the cigar
    assert result.cigar == Cigar("2M3M")
    assert result.q_start == 1


def test_fix_c_terminus_alignment():
    # given alignment of query and target
    alignment = AlignmentEntryAA(
        target_id="target_id",
        alignment_score=None,
        seq_identity=None,
        e_value=None,
        q_start=1,
        q_end=4,
        t_start=1,
        t_end=4,
        cigar="3M",
        species=None,
        locus=None,
        q_seq=None,
        t_seq=None,
    )
    query_sequence = "BAAAA"
    target_sequence = "CAAABBB"
    # alignment:
    # q:    B-AAAA
    # t:    -CAAABBB
    # aln:  SNMMM...

    # when
    result = force_c_terminus_matches(query_sequence, target_sequence, alignment)

    # then
    # verify N nor S was not appended to cigar
    # verify that min(q_start, t_start) matches was appended to the cigar
    assert result.cigar == Cigar("3M1M")


def test_merge_cigars():
    # I M -> I: it1++ -ok
    # I D -> M:it1++, it2++ -ok
    # I I -> I: it2++ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # M M ->M:it1++, it2++ -ok
    # M D -> D: it2++ -ok
    # M I -> I: it1++, it2++ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # D D -> D: it1++ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # D M -> : D : it1++, it2++
    # D I -> X: it1++, it2++

    # M M
    assert _merge_cigars("MMMM", "MMMM") == "MMMM"
    # M D
    assert _merge_cigars("MMM", "MMDM") == "MMDM"
    assert _merge_cigars("MMM", "MMDDM") == "MMDDM"
    # M I
    assert _merge_cigars("MMMM", "MMIM") == "MMIM"
    assert _merge_cigars("MMMM", "MMIIM") == "MMII"

    # I M
    assert _merge_cigars("MMIM", "MMMM") == "MMIM"
    assert _merge_cigars("MMIIM", "MMMMD") == "MMIIM"
    # I D
    assert _merge_cigars("MMIM", "MMDM") == "MMMM"
    assert _merge_cigars("MMIIM", "MMDDM") == "MMMMM"
    # I I
    assert _merge_cigars("MMIM", "MMIM") == "MMII"
    assert _merge_cigars("MMIIM", "MMIM") == "MMIII"
    assert _merge_cigars("MMIM", "MMIIM") == "MMII"
    assert _merge_cigars("MMIIM", "MMIIM") == "MMIII"

    # D M
    assert _merge_cigars("MMDM", "MMMM") == "MMDM"
    assert _merge_cigars("MMDDM", "MMMMM") == "MMDDM"
    # D D
    with pytest.raises(AssertionError, match="Query gene alignment is longer than gene scheme alignment"):
        _merge_cigars("MMDM", "MMDM") == "MMDDM"

    assert _merge_cigars("MMDM", "MMDDMM") == "MMDDDM"
    assert _merge_cigars("MMDM", "MMDMDM") == "MMDDDM"


def test_merge_long_cigars():
    q_g = "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNMMMMMMMMMMMIMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMIIIMMMMMMMMMMMMMM"
    # QVQLVESGGGLVKPGGSLRLSCAASGFTFSDYYMSWIRQAPGKGLEWVSYIS SSGSTIYYADSVKGRFTISRDNAKNSLYLQM   NSLRAEDTAVYYCAR
    g_sch = "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMIIMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMIIIMMMMMMMMMMMM"

    #        NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNMMMMMMMMMMMIIMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMIIIIIIMMMMMMMMMMMMMMMMMMMIIIIIIIIMMMMMMMMMMMM

    res = _merge_cigars(q_g, g_sch)
    print(res)


def test_infer_last_scheme_position_aligned():
    # given
    query = "CAAAAAAAAAAAA"
    gene = "BAAAAAAABC"
    aln = "SNMMMMMMM"
    g_sch = "MMMDMMMMMM"
    q_sch = "SNMMDMMMMM"

    # query =  "C-AA-AAAAAAAAAA"
    # gene =   "-BAA-AAAAABC"
    # aln =    "SNMM-MMMMM"
    # g_sch =   "MMMDMMMMMM"
    # q_sch2 = "IDMMDMMMMM"
    # last_pos="         |"  # 9 (0-based)

    # when
    last_pos = infer_last_scheme_position_aligned(q_sch)

    # then
    assert last_pos == 9


def test_merge_v_j():
    # given
    query_imgt_cigar_v = AlignmentString(
        "MMMMMMMMMDMMMMMMMMMMMMMMMMMMMMDDDDMMMMMMMMMMMMMMMMMMMMMMMMMDDDMMMMMMMMMMDMMMMMMMMMMMMMMMMMMDMMMMMMMMMMMMMM"
    )  # pylint: disable=line-too-long
    unaligned_middle_len = 18
    query_imgt_cigar_j = AlignmentString(unaligned_middle_len * "S" + "DMMMMMMMMMMM")
    j_gene_start_on_scheme = 117
    # when
    result = merge_v_j(query_imgt_cigar_v, query_imgt_cigar_j, j_gene_start_on_scheme)

    # then
    assert result == AlignmentString(
        "MMMMMMMMMDMMMMMMMMMMMMMMMMMMMMDDDDMMMMMMMMMMMMMMMMMMMMMMMMMDDDMMMMMMMMMMDMMMMMMMMMMMMMMMMMMDMMMMMMMMMMMMMMMMMMMMIIIIIIIMMMMMMMMMMMMMMMM"
    )


def test_merge_v_j_where_j_starts_on_deletion():
    # given
    # query:
    # genes
    #
    # gene_imgt

    # query_imgt

    query = "AAAAAAAAAABBB DDDDDDDDB"  # query is masked
    gene = "CCC DDDDDDDD"
    aln = "             NNN MMMMMMMM"
    g_sch = "MMMDMMMMMMMM"  # NNNNNNNNNNNNNNN -15N -> gene_start = 15
    q_sch = "DDDDMMMMMM"
    gene_start_on_scheme = 15

    query_imgt_cigar_v = AlignmentString(
        "MMMMMMMMMDMMMMMMMMMMMMMMMMMMMMDDDDMMMMMMMMMMMMMMMMMMMMMMMMMDDDMMMMMMMMMMDMMMMMMMMMMMMMMMMMMDMMMMMMMMMMMMMM"
    )  # pylint: disable=line-too-long
    unaligned_middle_len = 18
    query_imgt_cigar_j = AlignmentString(unaligned_middle_len * "S" + "DMMMMMMMMMMMMMM")
    j_gene_start_on_scheme = 114

    # when
    result = merge_v_j(query_imgt_cigar_v, query_imgt_cigar_j, j_gene_start_on_scheme)

    # then
    assert result == AlignmentString(
        "MMMMMMMMMDMMMMMMMMMMMMMMMMMMMMDDDDMMMMMMMMMMMMMMMMMMMMMMMMMDDDMMMMMMMMMMDMMMMMMMMMMMMMMMMMMDMMMMMMMMMMMMMMMMMMIIIIIIIIIIMMMMMMMMMMMMMMMMMM"
    )


def test_reorder_buffer_matches():
    # given buffer = cdr + 2 neigbour residues on both sides
    buffer = AlignmentString("MMMMMMMM")

    # when
    result = reorder_buffer_other_schemes(buffer, relative_indel_position=5)

    # then
    assert result == buffer


def test_reorder_buffer_insertions():
    # given buffer = cdr + 2 neigbour residues on both sides
    buffer = AlignmentString("MIMMMMMMMI")

    # when
    result = reorder_buffer_other_schemes(buffer, relative_indel_position=5)

    # then
    assert result == AlignmentString("MMMMMIIMMM")


def test_reorder_buffer_deletions():
    # given buffer = cdr + 2 neigbour residues on both sides
    buffer = AlignmentString("MDMMMMMMDM")

    # when
    result = reorder_buffer_other_schemes(buffer, relative_indel_position=5)

    # then
    assert result == AlignmentString("MMMDDMMMMM")


def test_reorder_buffer_imgt_matches():
    # given even len(buffer) - cdr1 and cdr2
    buffer = AlignmentString("MMMMMMMM")
    assert len(buffer) % 2 == 0

    # when
    result = reorder_buffer_imgt(buffer, relative_indel_position=5)

    # then
    assert result == buffer

    # given odd len(buffer) - cdr3
    buffer = AlignmentString("MMMMMMMMM")
    assert len(buffer) % 2 == 1

    # when
    result = reorder_buffer_imgt(buffer, relative_indel_position=5)

    # then
    assert result == buffer


def test_reorder_buffer_imgt_insertions():
    # given even len(buffer) - cdr1 and cdr2
    buffer = AlignmentString("MIMMMMMMMI")
    assert len(buffer) % 2 == 0

    # when
    result = reorder_buffer_imgt(buffer, relative_indel_position=5)

    # then
    assert result == AlignmentString("MMMMMIIMMM")

    # given odd len(buffer) - cdr3
    buffer = AlignmentString("MIMIIIMMMMMMI")
    assert len(buffer) % 2 == 1

    # when
    result = reorder_buffer_imgt(buffer, relative_indel_position=5)

    # then
    assert result == AlignmentString("MMMMMIIIIIMMM")


def test_reorder_buffer_imgt_deletions():
    # given even len(buffer) - cdr1 and cdr2
    buffer = AlignmentString("MDMMMMMMDM")
    assert len(buffer) % 2 == 0

    # when
    result = reorder_buffer_imgt(buffer, relative_indel_position=5)

    # then
    assert result == AlignmentString("MMMMDDMMMM")

    # given odd len(buffer) - cdr3
    buffer = AlignmentString("MDMMMMMMDMM")
    assert len(buffer) % 2 == 1

    # when
    result = reorder_buffer_imgt(buffer, relative_indel_position=5)

    # then
    assert result == AlignmentString("MMMMDDMMMMM")


def test_fix_imgt_cdrs_numbering():
    # CDR1
    # No change
    numbering = {32: [("A", 0), ("B", 1)], 33: [("D", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {32: [("A", 0), ("B", 1)], 33: [("D", 0)]}

    # Odd insertion number
    numbering = {32: [("A", 0), ("B", 1), ("C", 2), ("D", 3)], 33: [("E", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {32: [("A", 0), ("B", 1), ("C", 2)], 33: [("E", 0), ("D", 1)]}

    # Even insertion number
    numbering = {32: [("A", 0), ("B", 1), ("C", 2), ("D", 3), ("E", 4)], 33: [("F", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {
        32: [("A", 0), ("B", 1), ("C", 2)],
        33: [("F", 0), ("E", 1), ("D", 2)],
    }

    # CDR2
    # No change
    numbering = {60: [("A", 0), ("B", 1)], 61: [("D", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {60: [("A", 0), ("B", 1)], 61: [("D", 0)]}

    # Odd insertion number
    numbering = {60: [("A", 0), ("B", 1), ("C", 2), ("D", 3)], 61: [("E", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {60: [("A", 0), ("B", 1), ("C", 2)], 61: [("E", 0), ("D", 1)]}

    # Even insertion number
    numbering = {60: [("A", 0), ("B", 1), ("C", 2), ("D", 3), ("E", 4)], 61: [("F", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {
        60: [("A", 0), ("B", 1), ("C", 2)],
        61: [("F", 0), ("E", 1), ("D", 2)],
    }

    # CDR3 - swap priority
    # No change
    numbering = {111: [("A", 0)], 112: [("B", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {111: [("A", 0)], 112: [("B", 0)]}

    # Odd insertion number
    numbering = {111: [("A", 0), ("B", 1)], 112: [("C", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {111: [("A", 0)], 112: [("C", 0), ("B", 1)]}

    # Even insertion number
    numbering = {111: [("A", 0), ("B", 1), ("C", 2)], 112: [("D", 0)]}
    assert fix_imgt_cdrs_numbering(numbering) == {111: [("A", 0), ("B", 1)], 112: [("D", 0), ("C", 1)]}


if __name__ == "__main__":
    test_align_aa()
    test_force_n_terminus_alignments()
    test_fix_c_terminus_alignment()
    test_merge_cigars()

    test_infer_last_scheme_position_aligned()

    test_merge_v_j()
    test_merge_v_j_where_j_starts_on_deletion()

    test_reorder_buffer_matches()
    test_reorder_buffer_insertions()
    test_reorder_buffer_deletions()

    test_merge_long_cigars()



# Copyright 2016-2022 Paul Durivage
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def dict_merge(dct: dict, merge_dct: dict) -> dict:
    """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k in merge_dct.keys():
        if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], dict):  # noqa
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]
    return dct



from pathlib import Path

import pandas as pd


def melt_heavy_light(df: pd.DataFrame) -> pd.DataFrame:
    df = df[["therapeutic", "heavy_sequence", "light_sequence"]].melt(
        ["therapeutic"], var_name="chain_type", value_name="sequence"
    )
    df["sequence_id"] = df["therapeutic"] + "_" + df["chain_type"].str.strip("_sequence")
    df = df.set_index("sequence_id")
    df = df[df["sequence"] != "na"]
    return df


DIR = Path(__file__).parent.parent.parent / "data" / "therapeutics"

if __name__ == "__main__":
    sequence_types = ["human", "humanized"]
    dataset = pd.read_parquet(DIR / "therapeutics.parquet")
    dataset = dataset.drop_duplicates(subset=["therapeutic"]).reset_index(drop=True)

    dataset_heavy_light = melt_heavy_light(dataset)

    dataset_heavy_light.to_csv(DIR / "therapeutics.csv")

    with open(DIR / "therapeutics_aa.fasta", "w") as file:
        for pdb_id_chain, sequence in dataset_heavy_light[["sequence"]].itertuples(name=None):
            file.write(f">{pdb_id_chain}\n")
            file.write(f"{sequence}\n")

    metadata = pd.read_csv(DIR / "metadata.csv", skiprows=1)
    metadata.columns = metadata.columns.str.strip()
    metadata = metadata[["therapeutic", "type"]].copy()
    metadata["therapeutic"] = metadata["therapeutic"].str.strip().str.lower().str.capitalize()
    metadata["type"] = metadata["type"].str.strip().str.lower()
    metadata_human = metadata[metadata["type"].isin(sequence_types)]

    dataset_human = dataset.merge(metadata_human[["therapeutic"]], on="therapeutic")

    dataset_heavy_light_human = melt_heavy_light(dataset_human)
    dataset_heavy_light_human.to_csv(DIR / "therapeutics_human.csv")

    with open(DIR / "therapeutics_human_aa.fasta", "w") as file:
        for pdb_id_chain, sequence in dataset_heavy_light_human[["sequence"]].itertuples(name=None):
            file.write(f">{pdb_id_chain}\n")
            file.write(f"{sequence}\n")



import csv
from pathlib import Path
from typing import Optional

from riot_na.config import GENE_DB_DIR
from riot_na.data.model import AlignmentString, GeneId, Organism, Scheme


class SchemeMappingFacade:
    mappings: dict[GeneId, AlignmentString]

    def __init__(self, scheme: Scheme, allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR):
        if not allowed_species:
            allowed_species = [Organism.HOMO_SAPIENS, Organism.MUS_MUSCULUS]

        self.mappings = {}

        for species in allowed_species:
            file_path = db_dir / "scheme_mappings" / species.value / scheme.value / "scheme_mapping.csv"

            with file_path.open() as scheme_mapping_file:
                scheme_mapping_reader = csv.DictReader(scheme_mapping_file)
                for row in scheme_mapping_reader:
                    self.mappings[species.value + "|" + row["gene_id"]] = AlignmentString(row["scheme_cigar"])

    def get_mapping(self, organism: Organism, gene_id: str) -> AlignmentString:
        return AlignmentString(self.mappings[organism.value + "|" + gene_id])


if __name__ == "__main__":
    scheme_mapping_data_path = GENE_DB_DIR
    scheme_mapping_db = SchemeMappingFacade(
        scheme=Scheme.IMGT, allowed_species=[Organism.HOMO_SAPIENS], db_dir=scheme_mapping_data_path
    )

    print(scheme_mapping_db.mappings["IGHV1-2*07"])
    print(scheme_mapping_db.mappings["IGKJ4*01"])



from typing import Optional

from riot_na.airr.airr_validator import validate_airr_entry_aa
from riot_na.alignment.alignment_utils import (
    align_sequences,
    fold_cigar,
    offset_alignments,
    unfold_cigar,
)
from riot_na.data.model import (
    AirrRearrangementEntryAA,
    AlignmentEntryAA,
    AlignmentString,
    RegionOffsetsAA,
    Scheme,
    SchemeAlignment,
)


class AirrBuilderAA:  # pylint: disable=too-many-instance-attributes
    def __init__(self, sequence_header: str, sequence: str, scheme: Scheme):
        self.sequence_header = sequence_header
        self.sequence_aa = sequence
        self.scheme = scheme

        self.rearrangement = AirrRearrangementEntryAA(
            sequence_header=sequence_header, sequence_aa=sequence, numbering_scheme=scheme.value
        )

        self.v_gene_sequence_aa: Optional[str] = None
        self.v_gene_alignment_aa: Optional[AlignmentEntryAA] = None

        self.j_gene_alignment_aa: Optional[AlignmentEntryAA] = None
        self.j_gene_sequence_aa: Optional[str] = None

        self.scheme_alingment: Optional[SchemeAlignment] = None
        self.aa_offsets: Optional[RegionOffsetsAA] = None

    def with_v_gene_alignment_aa(self, v_aln: AlignmentEntryAA):
        self.v_gene_sequence_aa = v_aln.t_seq
        self.v_gene_alignment_aa = v_aln

        if v_aln is None:
            return self

        self.rearrangement.v_call = v_aln.target_id
        self.rearrangement.locus = v_aln.locus.value

        # alignment positions are 0-indexed, add 1 to start index to convert to 1-based
        self.rearrangement.v_sequence_start_aa = v_aln.q_start + 1
        self.rearrangement.v_sequence_end_aa = v_aln.q_end

        self.rearrangement.v_germline_start_aa = v_aln.t_start + 1
        self.rearrangement.v_germline_end_aa = v_aln.t_end

        v_sequence_segment = self.sequence_aa[v_aln.q_start : v_aln.q_end]
        v_germline_segment = v_aln.t_seq[v_aln.t_start : v_aln.t_end]

        v_sequence_alignment, v_germline_alignment = align_sequences(
            v_sequence_segment, v_germline_segment, unfold_cigar(v_aln.cigar)
        )

        self.rearrangement.v_sequence_alignment_aa = v_sequence_alignment
        self.rearrangement.v_germline_alignment_aa = v_germline_alignment

        # 1 based, hence need to add 1
        self.rearrangement.v_alignment_start_aa = 1
        self.rearrangement.v_alignment_end_aa = len(v_sequence_alignment)

        self.rearrangement.v_score_aa = v_aln.alignment_score
        self.rearrangement.v_cigar_aa = v_aln.cigar
        self.rearrangement.v_support_aa = v_aln.e_value
        self.rearrangement.v_identity_aa = v_aln.seq_identity

        return self

    def with_j_gene_alignment_aa(self, j_aln: AlignmentEntryAA):
        self.j_gene_sequence_aa = j_aln.t_seq

        if j_aln is not None:
            assert self.v_gene_alignment_aa is not None

            j_gene_alignment_aa = offset_alignments(self.v_gene_alignment_aa.q_end, j_aln)
            self.j_gene_alignment_aa = j_gene_alignment_aa

            self.rearrangement.j_call = j_gene_alignment_aa.target_id
            # alignment positions are 0-indexed, add 1 to start index to convert to 1-based
            self.rearrangement.j_sequence_start_aa = j_gene_alignment_aa.q_start + 1
            self.rearrangement.j_sequence_end_aa = j_gene_alignment_aa.q_end

            self.rearrangement.j_germline_start_aa = j_gene_alignment_aa.t_start + 1
            self.rearrangement.j_germline_end_aa = j_gene_alignment_aa.t_end

            j_sequence_segment = self.sequence_aa[j_gene_alignment_aa.q_start : j_gene_alignment_aa.q_end]
            j_germline_segment = j_aln.t_seq[j_aln.t_start : j_aln.t_end]

            j_sequence_alignment, j_germline_alignment = align_sequences(
                j_sequence_segment, j_germline_segment, unfold_cigar(j_aln.cigar)
            )

            self.rearrangement.j_sequence_alignment_aa = j_sequence_alignment
            self.rearrangement.j_germline_alignment_aa = j_germline_alignment

            # 1 based, hence need to add 1
            assert self.rearrangement.v_sequence_start_aa is not None
            v_alignment_str = unfold_cigar(self.v_gene_alignment_aa.cigar)
            deletions_on_v = v_alignment_str.count("D")
            self.rearrangement.j_alignment_start_aa = (
                j_gene_alignment_aa.q_start + 1 - self.v_gene_alignment_aa.q_start + deletions_on_v
            )
            self.rearrangement.j_alignment_end_aa = (
                j_gene_alignment_aa.q_end - self.v_gene_alignment_aa.q_start + deletions_on_v
            )

            self.rearrangement.j_score_aa = j_gene_alignment_aa.alignment_score
            self.rearrangement.j_cigar_aa = j_gene_alignment_aa.cigar
            self.rearrangement.j_support_aa = j_gene_alignment_aa.e_value
            self.rearrangement.j_identity_aa = j_gene_alignment_aa.seq_identity

        return self

    def _extract_region_aa(self, start: int, end: int):
        if start is None or end is None or start == -1 or end == -1:
            return None

        if self.rearrangement.sequence_aa:
            return self.rearrangement.sequence_aa[start - 1 : end]
        return None

    def with_aa_region_offsets(self, offsets: RegionOffsetsAA):
        self.aa_offsets = offsets

        self.rearrangement.fwr1_start_aa = offsets.fwr1_start_aa
        self.rearrangement.fwr1_end_aa = offsets.fwr1_end_aa
        self.rearrangement.cdr1_start_aa = offsets.cdr1_start_aa
        self.rearrangement.cdr1_end_aa = offsets.cdr1_end_aa
        self.rearrangement.fwr2_start_aa = offsets.fwr2_start_aa
        self.rearrangement.fwr2_end_aa = offsets.fwr2_end_aa
        self.rearrangement.cdr2_start_aa = offsets.cdr2_start_aa
        self.rearrangement.cdr2_end_aa = offsets.cdr2_end_aa
        self.rearrangement.fwr3_start_aa = offsets.fwr3_start_aa
        self.rearrangement.fwr3_end_aa = offsets.fwr3_end_aa
        self.rearrangement.cdr3_start_aa = offsets.cdr3_start_aa
        self.rearrangement.cdr3_end_aa = offsets.cdr3_end_aa
        self.rearrangement.fwr4_start_aa = offsets.fwr4_start_aa
        self.rearrangement.fwr4_end_aa = offsets.fwr4_end_aa

        self.rearrangement.fwr1_aa = self._extract_region_aa(offsets.fwr1_start_aa, offsets.fwr1_end_aa)
        self.rearrangement.cdr1_aa = self._extract_region_aa(offsets.cdr1_start_aa, offsets.cdr1_end_aa)
        self.rearrangement.fwr2_aa = self._extract_region_aa(offsets.fwr2_start_aa, offsets.fwr2_end_aa)
        self.rearrangement.cdr2_aa = self._extract_region_aa(offsets.cdr2_start_aa, offsets.cdr2_end_aa)
        self.rearrangement.fwr3_aa = self._extract_region_aa(offsets.fwr3_start_aa, offsets.fwr3_end_aa)
        self.rearrangement.cdr3_aa = self._extract_region_aa(offsets.cdr3_start_aa, offsets.cdr3_end_aa)
        self.rearrangement.fwr4_aa = self._extract_region_aa(offsets.fwr4_start_aa, offsets.fwr4_end_aa)

        return self

    def with_aa_scheme_alignment(self, alignment: SchemeAlignment):
        self.rearrangement.sequence_aa_scheme_cigar = fold_cigar(
            AlignmentString(alignment.alignment_str.replace("S", "").replace("N", ""))
        )
        return self

    def with_scheme_residue_mapping(self, mapping: dict[str, str]):
        self.rearrangement.scheme_residue_mapping = mapping
        return self

    def with_positional_scheme_mapping(self, mapping: dict[int, str]):
        self.rearrangement.positional_scheme_mapping = mapping
        return self

    def with_scheme_alignment_exc(self, exc: str):
        self.rearrangement.exc = exc
        return self

    def build(self):  # pylint: disable=too-many-statements
        if self.v_gene_alignment_aa is None:  # or self.rearrangement.sequence_alignment_aa is None:
            return self.rearrangement

        alignment_end = self.j_gene_alignment_aa.q_end if self.j_gene_alignment_aa else self.v_gene_alignment_aa.q_end
        sequence_alignment_aa_segment = self.sequence_aa[self.v_gene_alignment_aa.q_start : alignment_end]

        v_germline_segment = self.v_gene_alignment_aa.t_seq[
            self.v_gene_alignment_aa.t_start : self.v_gene_alignment_aa.t_end
        ]
        j_germline_segment = (
            self.j_gene_alignment_aa.t_seq[self.j_gene_alignment_aa.t_start : self.j_gene_alignment_aa.t_end]
            if self.j_gene_alignment_aa
            else ""
        )

        germline_segment = v_germline_segment + j_germline_segment

        v_aln_str = unfold_cigar(self.v_gene_alignment_aa.cigar)
        j_aln_str = unfold_cigar(self.j_gene_alignment_aa.cigar) if self.j_gene_alignment_aa else ""

        vj_junction_len = (
            self.j_gene_alignment_aa.q_start - self.v_gene_alignment_aa.q_end if self.j_gene_alignment_aa else 0
        )
        vj_junction_alignment_str = vj_junction_len * "I"

        full_alignment_str = v_aln_str + vj_junction_alignment_str + j_aln_str

        sequence_alignment_aa, germline_alignment_aa = align_sequences(
            sequence_alignment_aa_segment, germline_segment, full_alignment_str
        )
        self.rearrangement.sequence_alignment_aa = sequence_alignment_aa
        self.rearrangement.germline_alignment_aa = germline_alignment_aa

        if self.rearrangement.cdr3_aa is not None:
            jun_start_aa = self.aa_offsets.cdr3_start_aa - 1
            jun_end_aa = self.aa_offsets.cdr3_end_aa + 1

            self.rearrangement.junction_aa = self._extract_region_aa(jun_start_aa, jun_end_aa)
            self.rearrangement.junction_aa_length = len(self.rearrangement.junction_aa or "")

        # annotate
        self.rearrangement.stop_codon = "*" in self.rearrangement.sequence_alignment_aa
        if self.aa_offsets:
            self.rearrangement.complete_vdj = (
                self.aa_offsets.fwr1_start_aa
                and self.aa_offsets.fwr1_start_aa != -1
                and self.aa_offsets.fwr4_end_aa
                and self.aa_offsets.fwr4_end_aa != -1
            )

        self.rearrangement.productive = bool(
            not self.rearrangement.stop_codon and self.rearrangement.v_call and self.rearrangement.j_call
        )

        self.rearrangement.additional_validation_flags = validate_airr_entry_aa(self.rearrangement, scheme=self.scheme)

        return self.rearrangement



from riot_na.data.model import AlignmentString, ChainType, Scheme
from riot_na.data.scheme_definitions import (
    MARTIN_POSITIONS_HEAVY,
    MARTIN_POSITIONS_LIGHT,
)
from riot_na.schemes.smooth_alignment import smooth_cdr_junctions

SCHEME = Scheme.MARTIN


def test_smooth_cdr_junctions_martin_insertions_and_deletions():
    regions = {
        "fwr1": "MMDMDMIIMMDMDDMMDMMDMMIIMMMMIMII",
        "cdr1": "DMDMIIMDM",
        "fwr2": "IMIMMDMDMMMDMMDDDMMMIMI",
        "cdr2": "MDDDM",
        "fwr3": "MDMMMMMDMMMMMMMMDMMMIIMMMMDMMIIMMMMIMMDMDMMIMIII",
        "cdr3": "MDMMMM",
        "fwr4": "IIMMMMDMMMDMMIIM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == MARTIN_POSITIONS_HEAVY

    regions = {
        "fwr1": "MMDMDMIIMMDMDDMMDMMDMMIIMMMMIM",
        "cdr1": "MMMMMMIIM",
        "fwr2": "MIMMDMDMMMDMMDDDMMMIM",
        "cdr2": "DDMMM",
        "fwr3": "MDMMMMMDMMMMMMMMDMMMIIMMMMDMMIIMMMMIMMDMDMMIM",
        "cdr3": "MMMMMIIIIM",
        "fwr4": "MMMMDMMMDMMIIM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == MARTIN_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=ChainType.HEAVY, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_martin_insertions_and_deletions_light_chain():
    regions = {
        "fwr1": "MMMMIMMMMDMMMMIIMMDDMMMMMMDMI",
        "cdr1": "MDMMIMMIMI",
        "fwr2": "MDMDMMMMIIMMDMMMMMD",
        "cdr2": "MDM",
        "fwr3": "IMIMMDDMMMMMIIMMMDMMMDMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MDMDDD",
        "fwr4": "MIMMMMIIMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == MARTIN_POSITIONS_LIGHT

    regions = {
        "fwr1": "MMMMIMMMMDMMMMIIMMDDMMMMMMDM",
        "cdr1": "MMMMMIIIMM",
        "fwr2": "MDMDMMMMIIMMDMMMMMD",
        "cdr2": "MMM",
        "fwr3": "MIMMDDMMMMMIIMMMDMMMDMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMDDM",
        "fwr4": "MIMMMMIIMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == MARTIN_POSITIONS_LIGHT

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=ChainType.LIGHT, scheme=SCHEME)
        == then_alignment_str
    )


if __name__ == "__main__":
    test_smooth_cdr_junctions_martin_insertions_and_deletions()
    test_smooth_cdr_junctions_martin_insertions_and_deletions_light_chain()



import json

import pandas as pd

from riot_na.common.serialization_utils import base64_decode


def base64_decode_series(df, col):
    df = df[df[col].notna()].copy()
    df[col] = df[col].apply(base64_decode).apply(json.loads)
    return df


def not_na_df(df):
    return df[(df.ne("") & df.notna()).all(axis=1)]


def calculate_validation_flags_summary(df):
    df_notna = df[df["v_call"].notna() & (df["v_call"] != "") & df["productive"]]
    return df_notna["additional_validation_flags"].apply(pd.Series).apply(lambda x: x[x.notna()].value_counts()).T


def validation_flags_comparison(df_1, df_2, df_1_name: str = "old", df_2_name: str = "new"):
    df_1_notna = df_1[df_1["v_call"].notna() & (df_1["v_call"] != "") & df_1["productive"]]
    df_2_notna = df_2[df_2["v_call"].notna() & (df_2["v_call"] != "") & df_2["productive"]]
    common_index = df_1_notna.index.intersection(df_2_notna.index)
    df_1_val_flags = (
        df_1_notna.loc[common_index]["additional_validation_flags"]
        .apply(pd.Series)
        .apply(lambda x: x[x.notna()].value_counts())
        .T
    )
    df_2_val_flags = (
        df_2_notna.loc[common_index]["additional_validation_flags"]
        .apply(pd.Series)
        .apply(lambda x: x[x.notna()].value_counts())
        .T
    )
    comp = df_1_val_flags.join(df_2_val_flags, lsuffix=f"_{df_1_name}", rsuffix=f"_{df_2_name}")
    comp[f"False_{df_1_name}-{df_2_name}"] = comp[f"False_{df_1_name}"] - comp[f"False_{df_2_name}"]
    comp[f"True_{df_1_name}-{df_2_name}"] = comp[f"True_{df_1_name}"] - comp[f"True_{df_2_name}"]
    return comp[
        [
            f"False_{df_1_name}",
            f"False_{df_2_name}",
            f"False_{df_1_name}-{df_2_name}",
            f"True_{df_1_name}",
            f"True_{df_2_name}",
            f"True_{df_1_name}-{df_2_name}",
        ]
    ]



from riot_na.data.model import AlignmentString, ChainType, Scheme
from riot_na.data.scheme_definitions import KABAT_POSITIONS_HEAVY, KABAT_POSITIONS_LIGHT
from riot_na.schemes.smooth_alignment import (
    _reorder_cdr_deletions,
    smooth_cdr_junctions,
)

SCHEME = Scheme.KABAT
CHAIN_TYPE = ChainType.HEAVY


def test_smooth_cdr_junctions_kabat_no_indels():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMM",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == given_alignment_str
    )


def test_smooth_cdr_junctions_kabat_deletions_in_frameworks():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMD",
        "cdr1": "MMMMM",
        "fwr2": "DMMMMMMMMMMMMD",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MDMMMMMMMMMMMMMMMMMMMMMMMMMMD",
        "cdr3": "MMMMMMMM",
        "fwr4": "DMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    then_alignment_str = given_alignment_str

    temp = smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
    print(temp)
    assert temp == then_alignment_str


def test_smooth_cdr_junctions_kabat_deletions_in_cdrs():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMD",
        "cdr1": "DDMMM",
        "fwr2": "DMMMMMMMMMMMMD",
        "cdr2": "MMMMMMMDMMMDMMMM",
        "fwr3": "MDMMMMMMMMMMMMMMMMMMMMMMMMMMD",
        "cdr3": "DMMMMMDM",
        "fwr4": "DMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMD",
        "cdr1": "MMMDD",
        "fwr2": "DMMMMMMMMMMMMD",
        "cdr2": "MDDMMMMMMMMMMMMM",
        "fwr3": "MDMMMMMMMMMMMMMMMMMMMMMMMMMMD",
        "cdr3": "MMMMDDMM",
        "fwr4": "DMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    temp = smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
    assert temp == then_alignment_str


def test_smooth_cdr_junctions_kabat_too_much_deletions():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "DDMDD",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMDDMDMMD",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DMDDDMDD",
        "fwr4": "MMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MDDDD",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "DDDDMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DDDDDDMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_insertions_in_frameworks():
    regions = {
        "fwr1": "MMMMMIIMMMMIMMMMMMMIMMMMMMMMMMMMMIM",
        "cdr1": "MMMMM",
        "fwr2": "MIMMMIMMMIMMMMIMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MIMMIIMMMMMMMMMMMMMMMMMIIMMMMMMMMIM",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMMIMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    then_alignment_str = given_alignment_str
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_insertions_in_cdrs():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMII",
        "cdr1": "MIMMIMMII",
        "fwr2": "IIMMMMMMMMMMMMMMII",
        "cdr2": "MIIMMMMMMIIMMMMMMMMIMI",
        "fwr3": "IMMMMMMMMMMMMMMMMMMMMMMMMMMMMMI",
        "cdr3": "IMMMMIMMMIIM",
        "fwr4": "IMMMMMMMMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMMIIIIIIII",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMIIIIIIIIIMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMIIIIIIMM",
        "fwr4": "MMMMMMMMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_insertions_and_deletions():
    regions = {
        "fwr1": "MMMMIMMMMDMMMMIIMMDDMMMMMMMMMMMDMI",
        "cdr1": "MDMIMIMI",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMDMMMIMMIMDMMMMIMM",
        "fwr3": "IMIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MDMMMMMM",
        "fwr4": "MIMMMMIIMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "MMMMIMMMMDMMMMIIMMDDMMMMMMMMMMMDM",
        "cdr1": "MMMMMIII",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMMIIMMMMMMMMMMMMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIMM",
        "fwr4": "MIMMMMIIMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_start_on_last_fwr1():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNM",
        "cdr1": "MDMIMIMI",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMDMMMIMMIMDMMMMIMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNM",
        "cdr1": "MMMMMII",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMMIMMMMMMMMMMMMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIIMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_start_on_cdr1_start():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MDMIMIMI",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMDMMMIMMIMDMMMMIMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MMMMMII",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMMIMMMMMMMMMMMMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIIMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_start_on_cdr1():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNIMIM",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMDMMMIMMIMDMMMMIMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MMMMD",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMMIMMMMMMMMMMMMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIIMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_start_on_cdr1_end():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNM",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMDMMMIMMIMDMMMMIMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "MDDDD",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMMIMMMMMMMMMMMMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIIMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_start_on_fwr2():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNN",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMDMMMIMMIMDMMMMIMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNN",
        "fwr2": "MDMDMMMIIMMDMMMD",
        "cdr2": "MMMIMMMMMMMMMMMMM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIIMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_start_on_cdr2_end():
    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNN",
        "fwr2": "NNNNNNNNNNNNNN",
        "cdr2": "NNNNNNNNNNNNNNNM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMMII",
        "cdr3": "MMMMMMMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    regions = {
        "fwr1": "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
        "cdr1": "NNNNN",
        "fwr2": "NNNNNNNNNNNNNN",
        "cdr2": "DDDDDDDDDDDDDDDM",
        "fwr3": "MIMMMMIIMMMMMMMMMIIIMMMMDMMMMDMMMMM",
        "cdr3": "MMMMMMIIMM",
        "fwr4": "MIMMMIIMMMDMDM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_HEAVY

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_short_fwr4():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMM",
        "fwr4": "MMMMM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 107

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMM",
        "fwr4": "MMMMM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 107

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_fwr4_junction():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMM",
        "fwr4": "DM",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 104

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMM",
        "fwr4": "DM",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 104

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_in_fwr4_junction():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MDMMIMMMM",
        "fwr4": "M",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 103

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMMMM",
        "fwr4": "M",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 103

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_no_fwr4():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DMMMMMMM",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 102

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMMMDMM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 102

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_cdr3():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DMMMMM",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 100

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "MMMDDDMM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 102

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_cdr3_before_indel_pos():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DMM",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 97

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DDDDDDMM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 102

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_single_match_on_cdr3():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "M",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 95

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "DDDDDDDM",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 102

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_fwr3_junction():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 94

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 94

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_fwr3_junction_last_missing():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 93

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 93

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_fwr3():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 87

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMMMM",
        "fwr3": "MMMMMMMMMMMMMMMMMMMMMM",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 87

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_smooth_cdr_junctions_kabat_end_on_cdr2():
    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MMMMMMMMMMMMMDM",
        "fwr3": "",
        "cdr3": "",
        "fwr4": "",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == 64

    regions = {
        "fwr1": "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMM",
        "cdr1": "MMMMM",
        "fwr2": "MMMMMMMMMMMMMM",
        "cdr2": "MDDMMMMMMMMMMMMM",
        "fwr3": "",
        "cdr3": "",
        "fwr4": "",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == 65

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=CHAIN_TYPE, scheme=SCHEME)
        == then_alignment_str
    )


def test_reorder_buffer_deletions_kabat_no_matches():
    # even buffer len
    relative_indel_position = 6
    buffer = 14 * "D"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 15 * "D"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer


def test_reorder_buffer_deletions_kabat_single_match_at_buffer_end():
    # even buffer len
    relative_indel_position = 6
    buffer = 13 * "D" + "M"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 14 * "D" + "M"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer


def test_reorder_buffer_deletions_kabat_match_on_each_buffer_end():
    # even buffer len
    relative_indel_position = 6
    buffer = "M" + 13 * "D" + "M"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == 13 * "D" + "MM"

    # odd buffer len
    relative_indel_position = 7
    buffer = "M" + 14 * "D" + "M"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == 14 * "D" + "MM"


def test_reorder_buffer_deletions_kabat_two_matches_on_buffer_right_end():
    # even buffer len
    relative_indel_position = 6
    buffer = 12 * "D" + "MM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 13 * "D" + "MM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer


def test_reorder_buffer_deletions_kabat_three_matches_on_buffer_right_end():
    # even buffer len
    relative_indel_position = 6
    buffer = 11 * "D" + "MMM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer

    # odd buffer len
    relative_indel_position = 7
    buffer = 12 * "D" + "MMM"

    matches = buffer.count("M")
    deletions = buffer.count("D")

    assert _reorder_cdr_deletions(relative_indel_position, matches, deletions) == buffer


def test_smooth_cdr_junctions_kabat_insertions_and_deletions_light_chain():
    regions = {
        "fwr1": "MMMIMMMMDMMMMIMMMMIMMMMMMMII",
        "cdr1": "MDMMMIMMIMMDM",
        "fwr2": "DMMIMMMMDMMMMMMMI",
        "cdr2": "MMIMMIIMMD",
        "fwr3": "IMIMMMMMMMMDMMMMDMMMMMMMMMIMMMMMMMIMI",
        "cdr3": "IMDDMDMDMD",
        "fwr4": "IMIMMMMIMMDMMI",
    }
    given_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(given_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_LIGHT

    regions = {
        "fwr1": "MMMIMMMMDMMMMIMMMMIMMMMMMM",
        "cdr1": "MMMMIIMMMMMMM",
        "fwr2": "DMMIMMMMDMMMMMMM",
        "cdr2": "MMMIIIIMMMM",
        "fwr3": "MIMMMMMMMMDMMMMDMMMMMMMMMIMMMMMMMIM",
        "cdr3": "MMMMMDDMM",
        "fwr4": "MIMMMMIMMDMMI",
    }
    then_alignment_str = AlignmentString("".join(regions.values()))
    assert sum(map(then_alignment_str.count, ["M", "D", "N"])) == KABAT_POSITIONS_LIGHT

    assert (
        smooth_cdr_junctions(alignment_str=given_alignment_str, chain_type=ChainType.LIGHT, scheme=SCHEME)
        == then_alignment_str
    )


if __name__ == "__main__":
    test_smooth_cdr_junctions_kabat_no_indels()
    test_smooth_cdr_junctions_kabat_deletions_in_frameworks()
    test_smooth_cdr_junctions_kabat_deletions_in_cdrs()
    test_smooth_cdr_junctions_kabat_too_much_deletions()
    test_smooth_cdr_junctions_kabat_insertions_in_frameworks()
    test_smooth_cdr_junctions_kabat_insertions_in_cdrs()
    test_smooth_cdr_junctions_kabat_insertions_and_deletions()
    test_smooth_cdr_junctions_kabat_start_on_last_fwr1()
    test_smooth_cdr_junctions_kabat_start_on_cdr1_start()
    test_smooth_cdr_junctions_kabat_start_on_cdr1()
    test_smooth_cdr_junctions_kabat_start_on_cdr1_end()
    test_smooth_cdr_junctions_kabat_start_on_fwr2()
    test_smooth_cdr_junctions_kabat_start_on_cdr2_end()
    test_smooth_cdr_junctions_kabat_short_fwr4()
    test_smooth_cdr_junctions_kabat_end_on_fwr4_junction()
    test_smooth_cdr_junctions_kabat_end_in_fwr4_junction()
    test_smooth_cdr_junctions_kabat_no_fwr4()
    test_smooth_cdr_junctions_kabat_end_on_cdr3()
    test_smooth_cdr_junctions_kabat_end_on_cdr3_before_indel_pos()
    test_smooth_cdr_junctions_kabat_single_match_on_cdr3()
    test_smooth_cdr_junctions_kabat_end_on_fwr3_junction()
    test_smooth_cdr_junctions_kabat_end_on_fwr3_junction_last_missing()
    test_smooth_cdr_junctions_kabat_end_on_fwr3()
    test_smooth_cdr_junctions_kabat_end_on_cdr2()
    test_reorder_buffer_deletions_kabat_no_matches()
    test_reorder_buffer_deletions_kabat_single_match_at_buffer_end()
    test_reorder_buffer_deletions_kabat_match_on_each_buffer_end()
    test_reorder_buffer_deletions_kabat_two_matches_on_buffer_right_end()
    test_reorder_buffer_deletions_kabat_three_matches_on_buffer_right_end()
    test_smooth_cdr_junctions_kabat_insertions_and_deletions_light_chain()
